#ifndef NEURON_H
#define NEURON_H

#include <map>
#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <ctime>
#include <utility>
#include <cmath>
#include <vector>

using namespace std;

/** Classe Neur&ocirc;nio
  * @author Herinson B. Rodrigues \n
  * @p O neur&ocirc;nio artificial &eacute; uma estrutura l&oacute;gico-matem&aacute;tica que procura simular a forma,\n
  * o comportamento e as fun&ccedil;&otilde;es de um neur&ocirc;nio biol&oacute;gico. Assim sendo, os dendritos foram \n
  * substitu&iacute;dos por entradas, cujas liga&ccedil;&otilde;es com o corpo celular artificial s&atilde;o realizadas \n
  * atrav&eacute;s de elementos chamados de peso (simulando as sinapses). Os est&iacute;mulos captados pelas \n
  * entradas s&atilde;o processados pela fun&ccedil;&atilde;o de soma, e o limiar de disparo do neur&ocirc;nio biol&oacute;gico \n
  * foi substitu&iacute;do pela fun&ccedil;&atilde;o de transfer&ecirc;ncia. \n\n
  *
  * Representa&ccedil;&atilde;o do neur&ocirc;nio
  * @image html neuron.jpg
  */

class Neuron
{
    public:

        /** @var
         * Fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o
         */
        enum ActivationFunction {
            LINEAR,
            STEP,
            SIGMOID
        };
        /**
         * Construtor padr&atilde;o.
         */
        Neuron();

        /**
         * Construtor.
         * Construtor apenas com a fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o.
         * @param ActivationFunction Tipo da fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o
         */
        Neuron( ActivationFunction );

        /**
         * Construtor.
         * Inicializa o membro myInputs, ou seja, copia o vetor com as entradas
         * para o membro da clase
         * @param vector<double> Entradas do neur&ocirc;nio
         * @param ActivationFunction Tipo da fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o
         */
        Neuron( vector<double>, ActivationFunction );

        /**
         * Destrutor
         * Despacho de mem&oacute;ria
         */
        ~Neuron();

        /** Setter para o bias
        * Atribui ao membro privado bias um valor rand&ocirc;mico para o peso sin&aacute;ptico
        * @return void Nenhum
        */
        void setLoadBias();

        /** Copia o vector com as entradas(ou sa&iacute;das da camada anterior) para o neur&ocirc;nio.
        * A(s) entrada(s) (s&atilde;o) armazenada(s) no membro privado myInputs. Os pesos so aleatorios na primeira iterao
        * @param vector<double> inputs contendo as entradas para o neur&ocirc;nio
        * @return void Nenhum
        */
        void setInputs(vector<double>);

        /** Copia o vector com as entradas(ou sa&iacute;das da camada anterior) para o neur&ocirc;nio.
        * A(s) entrada(s) (s&atilde;o) armazenada(s) no membro privado myInputs. Os pesos no so aleatorios, pois, os mesmos encontran-se dispostos no segundo vector.
        * O segundo parmetro  utilizado no setter para o pesso do baias
        * @param vector<double> inputs contendo as entradas para o neur&ocirc;nio
        * @return void Nenhum
        */
        void setInputs(vector<double>, vector<double>, double);

        /** Preenche uma &uacute;nica entrada para o neur&ocirc;nio na primeira iterao. A entrada  repassada para o membro privado myInputs
        * @param double input contendo um n&uacute;mero com a entrada
        * @return void Nenhum
        */
        void setInput(double);

        /** Preenche uma &uacute;nica entrada e peso para o neur&ocirc;nio e armazena o valor no membro privado myInputs
        * @param double input n&uacute;mero contendo a entrada
        * @param double load n&uacute;mero contendo o peso da respectiva entrada
        * @return Nenhum
        */
        void setInput(double, double);

        /** Setter para o valor do erro do neur&ocirc;nio
        * @param double error
        * @return void Nenhum
        */
        void setError(double);

        /** Setter para o valor da sada esperada do neur&ocirc;nio
        * @param double expectedOutput
        * @return void Nenhum
        */
        void setExpectedOutput(double);

        /** Setter para o valor do delta do neur&ocirc;nio
        * @param double delta
        * @return void Nenhum
        */
        void setDelta(double);

        // --> Setters

        // <-- Getters

        /** Retorna a sa&iacute;da da fun&ccedil;&atilde;o soma
        * @param void Nenhum
        * @return double Valor da fun&ccedil;&atilde;o soma
        */
        double getValueSumFunction();

        /** Retorna a sa&iacute;da gerada para o neur&ocirc;nio
        * @param void Nenhum
        * @return double Sa&iacute;da do neur&ocirc;nio
        */
        double getOutput();

        /** Retorna o tipo da fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o (ex.: LINEAR, DEGRAU, SIGMOIDAL)
        * @param void Nenhum
        * @return ActivationFunction Tipo de fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o utilizada
        */
        ActivationFunction getActivationFunction();

        /** Retorna o valor do erro do neur&ocirc;nio
        * @param void Nenhum
        * @return const double
        */
        const double getError() const;

        /** Retorna o valor da sada esperada do neur&ocirc;nio
        * @param void Nenhum
        * @return const double
        */
        double getExpectedOutput() const;

        /** Retorna o valor do delta do neur&ocirc;nio
        * @param void Nenhum
        * @return const double
        */
        const double getDelta() const;

        // --> Getters

        /** Fun&ccedil;&atilde;o de soma
        * @param multimap<double, double> inputs Um multimap contendo as entradas e pesos
        * @return void Nenhum
        */
        void sumFunction(multimap<double, double>);

        /** Fun&ccedil;&atilde;o de soma
        * @param void Nenhum
        * @return void Nenhum
        */
        void sumFunction();

        // <-- Setters
        /** Setter para o tipo de fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o
        * @param ActivationFunction func Fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o
        * @return void Nenhum
        */
        void setActivationFunction(ActivationFunction);

        // <-- Activation methods
        /** Fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o - Degrau Sim&eacute;trico
        * Se o valor retornado pela fun&ccedil;&atilde;o soma for maior ou igual a zero (v >= 0)
        * ou retorna -1 se este valor for menor que zero (v < 0)
        * @param void Nenhum
        * @return double Resultado da fun&ccedil;&atilde;o ativa&ccedil;&atilde;o
        */
        double stepFunction();

        /** Fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o - Combina&ccedil;&atilde;o linear
        * Retorna o mesmo valor da fun&ccedil;&atilde;o soma
        * @param void Nenhum
        * @return double Resultado da fun&ccedil;&atilde;o ativa&ccedil;&atilde;o
        * @sa sumFunction()
        */
        double linearFunction();

        /** Fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o - Sigmoidal
        * Retorna a tangente hiperb&oacute;lica do resultado da fun&ccedil;&atilde;o soma
        * @param void Nenhum
        * @return double Resultado da fun&ccedil;&atilde;o ativa&ccedil;&atilde;o
        */
        double sigmoidFunction();

        // --> Activation methods

        // <-- Helpers

        /** M&eacute;todo para gerar n&uacute;meros aleat&oacute;rios de ponto flutuante
        * @param double low Limite inferior
        * @param double upp Limite superior
        * @return double Um n&uacute;mero de ponto flutuante aleat&oacute;rio
        */
        double randDouble(double, double);

        /** M&eacute;todo para transformar &acirc;ngulos em radianos para graus
        * @param double radian double Valor do &acirc;ngulo em radianos
        * @return double Valor do &acirc;ngulo em em graus
        */
        double radianToDegree(double);

        // --> Helpers

        void print();
        multimap<double, double> getMyInputs();
        void addBuffer(double);
        void startAssociation();
        void setBias(double);
        double deltaWeight(int);
        void changeInput(vector<double>);
        void changeWeight(int, double);
        double getXi(int);
        double getWi(int);
        double getMyLoadBias();

    private:
        /** @property
          * multimap para representar as entradas do neur&ocirc;nio e seus respectivos pesos
          * V&aacute;rias entradas (Xi) com um peso associado a cada uma (Pi),
          * alguns pesos com sinais excitat&oacute;rios (+) e outros com sinais inibit&oacute;rios (-).
          * Os valores de entrada e ativa&ccedil;&atilde;o das c&eacute;lulas podem ser discretos, nos conjuntos {0, 1} ou {-1, 0, 1} ou
          * cont&iacute;nuos nos intervalos [0,1] ou [-1,1].
          */
        multimap <double, double> myInputs;
        vector < pair <double,double> > MYINPUTS;

        /** @property
          * vetor usado para fazer a associaao entre uma entrada (Y) com uma respectiva chave(pesos sinpticos)
        */
        vector<double> myBuffer;

        /** @property
          * Resultado da soma obtida pela fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o
          */
        double myValueSumFunction;

        /** @property
          * Valor para o bias
          */
        double myLoadBias;

        /** @property
          * Fun&ccedil;&atilde;o de ativa&ccedil;&atilde;o a ser utilizada
          */
        ActivationFunction myActFunction;

        /** @property
          * Valor para o erro do neurnio
          */
        double myError;

        /** @property
          * Valor para a sada esperada do neurnio
          */
        double myExpectedOutput;

        /** @property
          * Valor para o delta do neurnio
          */
        double myDelta;
};

#endif // NEURON_H
